﻿
using System;
using System.Drawing;
using System.Collections.Generic;
using System.Data;
using SketchPad.Spatial;
using System.Data.EntityClient;
using System.Linq;
using System.Data.Objects;

namespace SketchPad.EF
{
    public class EFDb : SpatialDb
    {
        /// <summary>
        /// The LINQ to Entities DataContext
        /// </summary>
        private Entities _context;

        /// <summary>
        /// Callback invoked for visualizing index construction 
        /// </summary>
        private DisplayBoundingBox _displayBoundingBox;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="connectionString">Connection string</param>
        public EFDb(string connectionString)
        {
            this._context = new Entities(connectionString);
            this._displayBoundingBox = null;
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="connectionString">Connection string</param>
        /// <param name="displayBoundingBox">Callback invoked during index construction; can be null</param>
        public EFDb(string connectionString, DisplayBoundingBox displayBoundingBox)
        {
            this._context = new Entities(connectionString);
            this._displayBoundingBox = displayBoundingBox;
        }

        /// <summary>
        /// A delegate that can be invoked during the construction of the spatial index
        /// </summary>
        public override DisplayBoundingBox DisplayBoundingBoxDelegate
        {
            get
            {
                return this._displayBoundingBox;
            }
            set
            {
                this._displayBoundingBox = value;
            }
        }

        /// <summary>
        /// Create a line entity
        /// </summary>
        /// <param name="A">Start point of the line</param>
        /// <param name="B">End point of the line</param>
        /// <param name="description">Descriptive text associated with the shape</param>
        /// <returns>An instance of Line</returns>
        public override ILine CreateLine(Point A, Point B, string description)
        {
            Line line = Line.CreateLine(0, A.X, A.Y, B.X, B.Y);
            line.Description = description;
            this._context.AddToShape(line);

            return line;
        }

        /// <summary>
        /// Create a circle entity
        /// </summary>
        /// <param name="center">The center of the circle</param>
        /// <param name="circumference">A point on the circumference of the circle</param>
        /// <param name="description">Descriptive text associated with the shape</param>
        /// <returns>An instance of Circle</returns>
        public override ICircle CreateCircle(Point center, Point circumference, string description)
        {
            Circle circle = Circle.CreateCircle(0, center.X, center.Y, circumference.X, circumference.Y);
            circle.Description = description;
            this._context.AddToShape(circle);

            return circle;
        }

        /// <summary>
        /// Compute the area of the specified circle
        /// </summary>
        /// <param name="shape">The circle as an IDrawable instance</param>
        /// <returns>The area of the circle as a double value</returns>
        public override double ComputeCircleArea(IDrawable shape)
        {
            if (shape is Line)
            {
                return 0;
            }

            int id = ((Circle)shape).ID;
            object oResult = null;

            EntityCommand cmd = (EntityCommand)this._context.Connection.CreateCommand();
            cmd.CommandText = "Entities.ComputeCircleArea";
            cmd.CommandType = CommandType.StoredProcedure;
            
            cmd.Parameters.AddWithValue("circleID", id);

            EntityParameter param = cmd.CreateParameter();
            param.DbType = DbType.Decimal;
            param.Direction = ParameterDirection.Output;
            param.ParameterName = "area";
            cmd.Parameters.Add(param);

            bool closeConnection = false;
            try
            {
                if (cmd.Connection.State != ConnectionState.Open)
                {
                    closeConnection = true;
                    cmd.Connection.Open();
                }
                cmd.ExecuteNonQuery();
            }
            finally
            {
                if (closeConnection)
                    cmd.Connection.Close();
            }

            oResult = param.Value;

            if (oResult != null && !oResult.Equals(DBNull.Value))
            {
                return Convert.ToDouble(oResult);
            }

            return 0;
        }

        /// <summary>
        /// Remove a shape from the database
        /// </summary>
        /// <param name="shape">The shape to remove</param>
        public override void DeleteShape(IDrawable entity)
        {
            System.Diagnostics.Debug.Assert(entity != null);

            this._context.DeleteObject(entity as Shape);
        }

        public override void SaveChanges()
        {
            this._context.SaveChanges();
        }

        public override IList<IDrawable> GetAllShapes()
        {
            ObjectQuery<Shape> shapes = this._context.CreateQuery<Shape>(@"SELECT VALUE s FROM Shape AS s");
            var query = from s in shapes
                        select s;

            return query.AsEnumerable().Select(s => s).Cast<IDrawable>().ToList();
        }

        public override IList<IDrawable> FindShapes(string description)
        {
            List<IDrawable> result = new List<IDrawable>();
            if (string.IsNullOrEmpty(description))
            {
                return result;
            }
            
            string eSQL = string.Format(
                            @"SELECT VALUE s FROM Shape AS s where s.Description LIKE '%{0}%'",
                            description);

            ObjectQuery<Shape> shapes = this._context.CreateQuery<Shape>(eSQL);
            var query = from s in shapes
                        select s;

            return query.AsEnumerable().Select(s => s).Cast<IDrawable>().ToList();
        }

        public override string GetAllDescriptions()
        {
            ObjectQuery<string> shapeDescriptions = this._context.CreateQuery<String>(@"SELECT VALUE s.Description FROM Shape AS s");
            List<String> descriptions = (from s in shapeDescriptions
                                         select s).ToList();

            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            foreach (string s in descriptions)
                sb.AppendFormat("{0} ", s);

            return sb.ToString();
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (this._context != null)
                {
                    this._context.Dispose();
                    this._context = null;
                }
            }
        }
    }
}
