﻿
using System;
using System.Drawing;
using System.Collections.Generic;
using System.Data.Linq;
using System.Linq;
using SketchPad.Spatial;
using System.Data.Linq.Mapping;
using System.Reflection;


namespace SketchPad.LinqToSql
{
    public class LinqToSqlDb : SpatialDb
    {
        /// <summary>
        /// The LINQ to SQL DataContext
        /// </summary>
        private Context _context;

        /// <summary>
        /// Callback invoked for visualizing index construction 
        /// </summary>
        private DisplayBoundingBox _displayBoundingBox;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="dbConnString">The DB connection string</param>
        public LinqToSqlDb(string dbConnString)
        {
            this._context = new Context(dbConnString);
            this._displayBoundingBox = null;
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="dbConnString">The DB connection string</param>
        /// <param name="displayBoundingBox">Callback invoked during index construction;
        /// can be null</param>
        public LinqToSqlDb(string dbConnString, DisplayBoundingBox displayBoundingBox)
        {
            this._context = new Context(dbConnString);
            this._displayBoundingBox = displayBoundingBox;
        }

        /// <summary>
        /// A delegate that can be invoked during the construction of the spatial index
        /// </summary>
        public override DisplayBoundingBox DisplayBoundingBoxDelegate
        {
            get
            {
                return this._displayBoundingBox;
            }
            set
            {
                this._displayBoundingBox = value;
            }
        }

        /// <summary>
        /// Create a line entity
        /// </summary>
        /// <param name="A">Start point of the line</param>
        /// <param name="B">End point of the line</param>
        /// <param name="description">Descriptive text associated with the shape</param>
        /// <returns>An instance of Line</returns>
        public override ILine CreateLine(Point A, Point B, string description)
        {
            Line line = new Line(A, B, description);
            this._context.Shapes.InsertOnSubmit(line);
            return line;
        }

        /// <summary>
        /// Create a circle entity
        /// </summary>
        /// <param name="center">The center of the circle</param>
        /// <param name="circumference">A point on the circumference of the circle</param>
        /// <param name="description">Descriptive text associated with the shape</param>
        /// <returns>An instance of Circle</returns>
        public override ICircle CreateCircle(Point center, Point circumference, string description)
        {
            Circle circle = new Circle(center, circumference, description);
            this._context.Shapes.InsertOnSubmit(circle);
            return circle;
        }

        /// <summary>
        /// Compute the area of the specified circle
        /// </summary>
        /// <param name="shape">The circle as an IDrawable instance</param>
        /// <returns>The area of the circle as a double value</returns>
        public override double ComputeCircleArea(IDrawable shape)
        {
            if (shape is Line)
            {
                return 0;
            }

            int? id = ((Circle)shape).ID;
            double? area = 0;
            this._context.ComputeCircleAreaHelper(id, ref area);

            return area == null ? 0 : area.Value;
        }

        /// <summary>
        /// Remove a shape from the database
        /// </summary>
        /// <param name="shape">The shape to remove</param>
        public override void DeleteShape(IDrawable shape)
        {
            System.Diagnostics.Debug.Assert(shape != null);
            this._context.Shapes.DeleteOnSubmit(shape as Shape);
        }

        public override void SaveChanges()
        {
            this._context.SubmitChanges();
        }

        public override IList<IDrawable> GetAllShapes()
        {
            return this._context.Shapes.Select(s => s).Cast<IDrawable>().ToList();
        }

        public override IList<IDrawable> FindShapes(string description)
        {
            var query = from s in this._context.Shapes
                        where System.Data.Linq.SqlClient.SqlMethods.Like(s.Description, "%"+description+"%")
                        select s;

            return query.Cast<IDrawable>().ToList();
        }

        public override string GetAllDescriptions()
        {
            var query = from s in this._context.Shapes
                        select s.Description;

            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            foreach (string s in query)
                sb.AppendFormat("{0} ", s);

            return sb.ToString();
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (this._context != null)
                {
                    this._context.Dispose();
                    this._context = null;
                }
            }
        }
    }

    internal class Context : DataContext
    {
        public Table<Shape> Shapes;
        public Context(string connectionStr) : base(connectionStr) { }
        public Context(System.Data.IDbConnection connection) : base(connection) { }

        [Function(Name = "dbo.sprocComputeCircleArea")]
        public int ComputeCircleAreaHelper(
            [Parameter(DbType = "Int")] System.Nullable<int> circleID, 
            [Parameter(DbType = "Float")] ref System.Nullable<double> area)
        {
            IExecuteResult result = this.ExecuteMethodCall(this, ((MethodInfo)(MethodInfo.GetCurrentMethod())), circleID, area);
            area = ((System.Nullable<double>)(result.GetParameterValue(1)));
            return ((int)(result.ReturnValue));
        }
    }
}
