﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using System.Configuration;
using SketchPad.Spatial;
using SketchPad.EF;
using SketchPad.LinqToSql;

namespace SketchPad
{
    public partial class MainForm : Form
    {
#region Private fields

        private Splash _splashScreen;
        private bool _usingEF;
        private SpatialDb _spatialDb;          // entry-point into spatial land
        private ISpatialIndex _index;

        private IList<IDrawable> _displayShapes;   // shapes present in the database
        private IList<IDrawable> _selectedShapes;
        private IList<IDrawable> _contextMenuShapes;
        private IList<IDrawable> _newShapes;        // shapes added by the user in the session
        private List<IDrawable> _indexBounds;

        private bool _drawingLine;
        private bool _drawingCircle;
        private bool _selecting;

        private bool _editingShape;
        private bool _showBoundingBoxes;
        private bool _animateIndex;

        private Point _startPoint;
        private Point _endPoint;
        private Point _lastLocation;

        private ContextMenu _contextMenu;
        private StatusBar   _statusBar;

#endregion Private fields

#region Statics

        private static string s_linqToSqlConnectionString;
        private static string s_EfConnectionString;

        static MainForm()
        {
            ConnectionStringSettingsCollection settings =
                ConfigurationManager.ConnectionStrings;

            if (settings != null)
            {
                s_linqToSqlConnectionString = settings["LinqToSql"].ConnectionString;
                s_EfConnectionString = settings["EF"].ConnectionString;
            }
        }

#endregion Statics

#region Constructor

        /// <summary>
        /// Default constructor
        /// </summary>
        public MainForm() : this(false)
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        public MainForm(bool usingEF)
        {
            this._usingEF = usingEF;

            // display the splash screen
            this._splashScreen = new Splash(usingEF);
            this._splashScreen.Owner = this;
            this._splashScreen.Show();

            // process the message queue - so that the splash screen can be drawn
            Application.DoEvents();

            // Disable the main form so that it is not drawn over the splash screen;
            // it is reenabled in MainForm Load event handler
            this.Enabled = false;

            if (this._usingEF)
            {
                this._spatialDb = new EF.EFDb(MainForm.s_EfConnectionString);
            }
            else
            {
                this._spatialDb = new LinqToSql.LinqToSqlDb(MainForm.s_linqToSqlConnectionString);
            }

            this._displayShapes  = new List<IDrawable>();
            this._selectedShapes = new List<IDrawable>();
            this._newShapes      = new List<IDrawable>();

            this._startPoint = new Point(0, 0);
            this._endPoint   = new Point(0, 0);
            this._lastLocation = new Point(0, 0);

            InitializeComponent();

            this.Text += (this._usingEF ? " - Using the Entity Framework" : " - Using LINQ to SQL");

            // initialize the context menu
            this._contextMenu = new ContextMenu();
            this._contextMenu.Popup += new EventHandler(OnContextMenuPopup);
            this._contextMenu.MenuItems.Add("&Select", new EventHandler(OnClickContextMenuSelect));
            this._contextMenu.MenuItems.Add("&Delete", new EventHandler(OnClickContextMenuDelete));
            this._contextMenu.MenuItems.Add("&Compute Area", new EventHandler(OnClickContextMenuComputeArea));
            this.ContextMenu = this._contextMenu;

            this._statusBar = new StatusBar();
            this._statusBar.ShowPanels = false;
            this._statusBar.Text = "Ready";
            this.Controls.Add(this._statusBar);
        }

#endregion Constructor

#region Form event handlers

        private void Form1_Load(object sender, EventArgs e)
        {
            System.Threading.Thread.Sleep(5000);    // let the splash screen have its day

            this.Enabled = true;
            this.Visible = true;
            this._splashScreen.Close();

            Cursor.Current = Cursors.Default;
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            Graphics g = e.Graphics;
            e.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

            // shapes already added to the collection will be drawn with a special pen:
            // one that is 5 pixels wide, purple in color, and partially transparent.
            //
            using (Pen pen = new Pen(Color.FromArgb(150, Color.Purple), 5))
            {
                foreach (IDrawable shape in this._displayShapes)
                {
                    shape.Draw(g, pen);
                }
            }

            // if we are interactively editing a shape, kick in the rubberbanding logic
            //
            if (!this._startPoint.Equals(new Point(0, 0)))
            {
                bool endPointSelected = !this._endPoint.Equals(new Point(0, 0));
                Point end = endPointSelected ? this._endPoint : this.PointToClient(Cursor.Position);

                using (Pen rubberBandingPen = new Pen(Color.Black, 1))
                {
                    rubberBandingPen.DashStyle = System.Drawing.Drawing2D.DashStyle.DashDotDot;

                    if (this._drawingLine)
                    {
                        g.DrawLine(rubberBandingPen, this._startPoint, end);
                    }
                    else
                    if (this._drawingCircle)
                    {
                        // draw a radial line as well as the circle
                        g.DrawLine(rubberBandingPen, this._startPoint, end);
                        Util.DrawCircleHelper(g, rubberBandingPen, this._startPoint, end);
                    }
                    else
                    if (this._selecting)
                    {
                        g.DrawRectangle(rubberBandingPen,
                                        this._startPoint.X, this._startPoint.Y, // upper-left corner
                                        Math.Abs(end.X - this._startPoint.X), 
                                        Math.Abs(end.Y-this._startPoint.Y));
                    }
                }

                if (endPointSelected)
                {
                    this._startPoint = new Point(0, 0);
                    this._endPoint = new Point(0, 0);
                }
            }

            if (_showBoundingBoxes)
            {
                using (Pen bbPen = new Pen(Color.FromArgb(150, Color.Blue), 1))
                {
                    foreach (IDrawable shape in this._displayShapes)
                    {
                        g.DrawRectangle(bbPen, shape.GetBoundingBox());
                    }
                }
            }

            // display the index hierarchy: use a different color for each index box
            //
            if (this._indexBounds != null && this._indexBounds.Count > 0)
            {
                Color[] colors = new Color[] {
                                     Color.Black, Color.BlueViolet, Color.Brown, Color.Crimson, Color.Green,
                                     Color.HotPink, Color.Indigo, Color.DarkGray, Color.LightBlue, Color.Plum,
                                     Color.MediumSeaGreen, Color.Orange, Color.Pink, Color.Yellow, Color.DarkSalmon,
                                     Color.Violet, Color.Wheat, Color.LightSalmon, Color.LightGray, Color.DarkOrange
                                 };

                int colorIndex = 0;
                int numColors = colors.Length;
                try
                {
                    Cursor.Current = Cursors.WaitCursor;
                    if (this._animateIndex)
                    {
                        this._indexBounds.Sort(new IndexComparer());
                    }

                    foreach (IDrawable shape in this._indexBounds)
                    {
                        Color color = colors[colorIndex++ % numColors];
                        using (Pen indexPen = new Pen(Color.FromArgb(150, color), 1))
                        {
                            shape.Draw(g, indexPen);
                        }

                        if (this._animateIndex)
                        {
                            System.Threading.Thread.Sleep(500); // index creation visualization aid
                        }
                    }
                }
                finally
                {
                    Cursor.Current = Cursors.Default;
                }

                this._animateIndex = false;
            }

            // if we have one or more entities in the select list, draw their handles
            if (this._selectedShapes.Count > 0)
            {
                using (Pen selectPen = new Pen(Color.Black, 1))
                using (Pen hiLightPen = new Pen(Color.GreenYellow, 4))
                {
                    foreach (IDrawable shape in this._selectedShapes)
                    {
                        shape.Draw(g, hiLightPen);
                        shape.Select(g, selectPen);
                    }
                }
            }
        }

        private void Form1_MouseClick(object sender, MouseEventArgs e)
        {
            if (this._editingShape)
            {
                this._startPoint = new Point(e.X, e.Y);
                this._editingShape = false;
                this._selectedShapes.Clear();
            }
            else
            {
                try
                {
                    this._endPoint = new Point(e.X, e.Y);
                    if (this._drawingLine)
                    {
                        string description = this.DisplayFindFormWithCaption(@"Edit Description");
                        if (description != null)    // null indicates cancellation of drawing operation
                        {
                            ILine line = this._spatialDb.CreateLine(this._startPoint, this._endPoint, description);
                            this._displayShapes.Add(line);
                            this._newShapes.Add(line);
                            // re-index
                            this.BuildIndex(false);
                        }
                    }
                    else
                    if (this._drawingCircle)
                    {
                        string description = this.DisplayFindFormWithCaption(@"Edit Description");
                        if (description != null)    // null indicates cancellation of drawing operation
                        {
                            ICircle circle = this._spatialDb.CreateCircle(this._startPoint, this._endPoint, description);
                            this._displayShapes.Add(circle);
                            this._newShapes.Add(circle);
                            // re-index
                            this.BuildIndex(false);
                        }
                    }
                    else
                    if (this._selecting && this._index != null)
                    {
                        // Use the index to select objects here
                        this._selectedShapes = this._index.GetSelectedShapes(new Rectangle(this._startPoint.X, this._startPoint.Y,
                                                                    Math.Abs(this._endPoint.X - this._startPoint.X),
                                                                    Math.Abs(this._endPoint.Y - this._startPoint.Y)));
                    }
                }
                finally
                {
                    // reset
                    if (this._drawingLine || this._drawingCircle || this._selecting)
                    {
                        Cursor.Current = Cursors.Default;
                    }
                    this._startPoint = new Point(0, 0);
                    this._endPoint = new Point(0, 0);
                    this._drawingLine = false;
                    this._drawingCircle = false;
                    this._selecting = false;
                }
            }

            this.Invalidate();
        }

        private void Form1_MouseMove(object sender, MouseEventArgs e)
        {
            bool movingRight = this._startPoint.X < e.X;
            bool movingDown  = this._startPoint.Y < e.Y;

            if (this._drawingLine || this._selecting)
            {
                int xInvalidRectangle = movingRight ? this._startPoint.X : e.X;
                int yInvalidRectangle = movingDown ? this._startPoint.Y : e.Y;

                int diffX = Math.Max( Math.Abs(e.X - this._startPoint.X), Math.Abs(this._lastLocation.X - this._startPoint.X) );
                int diffY = Math.Max( Math.Abs(e.Y - this._startPoint.Y), Math.Abs(this._lastLocation.Y - this._startPoint.Y) );

                this.Invalidate(new Rectangle(
                                    xInvalidRectangle, yInvalidRectangle,   // upper-left corner
                                    diffX + 1, diffY + 1));  // need the +1 for the pen width
            }
            else
            // invalidate a rectangular region that circumscribes a circle with the radial
            // distance computed between _startPoint and max(e, _lastPoint)
            if (this._drawingCircle || this._editingShape)
            {
                int distCurrent  = Util.ComputeDistance(this._startPoint, e.Location);
                int distPrevious = Util.ComputeDistance(this._startPoint, this._lastLocation);

                this.Invalidate(Util.ComputeBoundingRectangle(
                                            this._startPoint,
                                            new Point(this._startPoint.X + Math.Max(distCurrent, distPrevious) + 1,
                                                      this._startPoint.Y)
                                        ));
            }

            this._lastLocation = e.Location;
        }

        private void Form1_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Escape)
            {
                // reset
                this._selectedShapes.Clear();
                this._startPoint = new Point(0, 0);
                this._endPoint = new Point(0, 0);
                this._drawingLine = false;
                this._drawingCircle = false;
                this._selecting = false;
                Cursor.Current = Cursors.Default;

                this._statusBar.Text = "Ready";
                this.Invalidate();
            }
            else
            if (e.KeyCode == Keys.Delete)
            {
                DialogResult result = MessageBox.Show(
                                "Are you sure you want to delete the selected shapes?",
                                "Confirm Delete",
                                MessageBoxButtons.YesNo,
                                MessageBoxIcon.Warning,
                                MessageBoxDefaultButton.Button2);

                if (result == DialogResult.Yes)
                {
                    int countDeleted = this._selectedShapes.Count;
                    foreach (IDrawable shape in this._selectedShapes)
                    {
                        IDrawable entity = null;
                        if (shape is ILine) entity = (ILine)shape;
                        else entity = (ICircle)shape;

                        if (this._newShapes.Contains(entity))
                        {
                            this._newShapes.Remove(entity);
                            this._displayShapes.Remove(shape);
                        }
                        else
                        if (this._displayShapes.Contains(shape))
                        {
                            // delete from database
                            this._spatialDb.DeleteShape(shape);
                            this._displayShapes.Remove(shape);
                        }
                    }
                    this._statusBar.Text = countDeleted.ToString() + " objects deleted";
                }
                this._selectedShapes.Clear();
                this.Invalidate();
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this._newShapes.Count > 0)
            {
                DialogResult result = MessageBox.Show(
                                "There are unsaved changes in the session. Are you sure you want to quit?",
                                "Warning - Unsaved Changes",
                                MessageBoxButtons.YesNo,
                                MessageBoxIcon.Warning,
                                MessageBoxDefaultButton.Button2);

                if (result == DialogResult.No)
                {
                    e.Cancel = true;
                }
            }
        }

#endregion Form event handlers

#region ToolStripMenuItem Event Handlers

        private void lineToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.Cross;

            // line-drawing mode; turn off other flags
            this._drawingLine = true;
            this._drawingCircle = false;
            this._selecting = false;
            this._editingShape = true;
            this.Invalidate();
        }

        private void circleToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.Cross;

            // circle-drawing mode; turn off other flags
            this._drawingLine = false;
            this._drawingCircle = true;
            this._selecting = false;
            this._editingShape = true;
        }

        private void selectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.Cross;

            // selection mode; turn off other flags
            this._drawingLine = false;
            this._drawingCircle = false;
            this._selecting = true;
            this._editingShape = true;
        }

        private void selectAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this._selectedShapes.Clear();
            ((List<IDrawable>)this._selectedShapes).AddRange(this._displayShapes);
            if (this._newShapes.Count > 0)
            {
                foreach (IDrawable s in this._newShapes)
                {
                    this._selectedShapes.Add(s);
                }
            }
            this.Invalidate();
        }

        private void clearSelectionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this._selectedShapes.Clear();
            this.Invalidate();
        }

        /// <summary>
        /// File --> Open Database
        /// </summary>
        /// <param name="sender">Sender</param>
        /// <param name="e">Event arguments</param>
        private void openToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.WaitCursor;

            try
            {
                this._displayShapes.Clear();
                this._newShapes.Clear();
                this._displayShapes = this._spatialDb.GetAllShapes();
                this._statusBar.Text = this._usingEF ? s_EfConnectionString : s_linqToSqlConnectionString;
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }

            this.BuildIndex(false);
            this.Invalidate();
        }

        private void showBoundingBoxesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem menuItem = sender as ToolStripMenuItem;

            this._showBoundingBoxes = menuItem.CheckState == CheckState.Checked;
            this.Invalidate();
        }

        private void showIndexHierarchyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.WaitCursor;
            ToolStripMenuItem menuItem = sender as ToolStripMenuItem;

            try
            {
                if (menuItem.CheckState == CheckState.Checked)
                {
                    this.BuildIndex(true);
                }
                else
                {
                    this._indexBounds.Clear();
                    if (this._newShapes.Count > 0)
                    {
                        this.BuildIndex(false);
                    }
                    this._statusBar.Text = "Ready"; // reset status bar text
                }
            }
            finally
            {
                Cursor.Current = Cursors.Default;
                this._animateIndex = true;
                this.Invalidate();
            }
        }

        private void saveToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            if (this._newShapes.Count == 0)
            {
                return;
            }

            Cursor.Current = Cursors.WaitCursor;

            try
            {
                this._spatialDb.SaveChanges();
            }
            finally
            {
                Cursor.Current = Cursors.Default;
                this._newShapes.Clear();
                ((IDisposable)this._spatialDb).Dispose();
            }
        }

        private void exitToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// Find objects in the drawing with description that matches a specified pattern
        /// </summary>
        /// <param name="sender">Sender</param>
        /// <param name="e">Event arguments</param>
        /// <remarks>
        /// This is a demonstration of how we can get (location) information from the database
        /// and use it in the index structure to identify objects. In theory, we could simply
        /// add the shapes returned by the ObjectContext directly into the _selectedShapes list.
        /// However, that would imply creating new objects, distinct from the ones already in
        /// the index, which may not be acceptable in a real-world scenario.
        /// </remarks>
        private void findToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this._selectedShapes.Clear();

            if (this._displayShapes.Count == 0)
            {
                this._statusBar.Text = "No objects to search";
                return;
            }
            else
            {
                string findText = this.DisplayFindFormWithCaption(@"Find");

                if (string.IsNullOrEmpty(findText))
                {
                    return;
                }

                // examine the entities in the database
                IList<IDrawable> result = this._spatialDb.FindShapes(findText);
                if (result != null && result.Count > 0)
                {
                    foreach (IDrawable drawable in result)
                    {
                        int id = drawable is ILine ? ((ILine)drawable).ID : ((ICircle)drawable).ID;
                        IDrawable entity = this._index.GetShapeByLocation(drawable.GetBoundingBox().Location, id);
                        if (entity != null)
                            this._selectedShapes.Add(entity);
                    }
                }

                // now check and see if we have any unsaved objects with description that matches
                // the specified text
                if (this._newShapes.Count > 0)
                {
                    foreach (IDrawable shape in this._newShapes)
                    {
                        string description = ((IIdentity)shape).Description;
                        if (!string.IsNullOrEmpty(description) && description.Contains(findText))
                        {
                            this._selectedShapes.Add(shape);
                        }
                    }
                }

                if (this._selectedShapes.Count > 0)
                {
                    this._statusBar.Text = this._selectedShapes.Count.ToString() + " objects found";
                }
                else
                {
                    this._statusBar.Text = "No match";
                }
            }

            this.Invalidate();
        }

        private void allDescriptionsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                this._statusBar.Text = this._spatialDb.GetAllDescriptions();
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

#endregion ToolStripMenuItem Event Handlers

#region Index Construction and Visualization

        public void ShowBoundingBox(Rectangle rectangle)
        {
            this._indexBounds.Add(new IndexBox(rectangle));
        }

        /// <summary>
        /// A private helper method to build an index over all entities on the canvas,
        /// including those that haven't been persisted.
        /// </summary>
        /// <param name="displayBoxes">Indicates whether the bounding boxes should be displayed</param>
        private void BuildIndex(bool displayBoxes)
        {
            if (displayBoxes)
            {
                if (this._indexBounds == null)
                {
                    this._indexBounds = new List<IDrawable>();
                }
                else
                {
                    this._indexBounds.Clear();
                }

                this._spatialDb.DisplayBoundingBoxDelegate = ShowBoundingBox;
            }
            else
            {
                this._spatialDb.DisplayBoundingBoxDelegate = null;
            }

            // displayShapes include newShapes
            this._index = this._spatialDb.BuildIndex(this._displayShapes);

            if (displayBoxes)
            {
                this.DisplayIndexStatistics();
            }
        }

#endregion Index Construction and Visualization

#region Context Menu Handlers

        void OnContextMenuPopup(object sender, EventArgs e)
        {
            // Determine the shapes under the click point and then save them off for
            // use by the context menu item handlers

            if (this._index != null)
            {
                Point ptClick = this.PointToClient(Cursor.Position);
                this._contextMenuShapes = this._index.GetShapesUnderPoint(ptClick);
            }

            // ** Validation **
            //Rectangle r = new Rectangle(ptClick, new Size(2, 2));
            //IList<IDrawable> shapes2 = this._index.GetSelectedShapes(r);
            //System.Diagnostics.Debug.Assert(shapes.Count == shapes2.Count);

            if (this._contextMenuShapes != null && this._contextMenuShapes.Count > 0)
            {
                foreach (MenuItem item in this._contextMenu.MenuItems)
                    item.Visible = true;
            }
            else
            {
                foreach (MenuItem item in this._contextMenu.MenuItems)
                    item.Visible = false;
            }
        }

        private void OnClickContextMenuSelect(object sender, System.EventArgs e)
        {
            if (this._contextMenuShapes != null && this._contextMenuShapes.Count > 0)
            {
                this._selectedShapes = this._contextMenuShapes;
            }
            else
            {
                this._selectedShapes.Clear();
            }

            this.Invalidate();
        }

        private void OnClickContextMenuDelete(object sender, System.EventArgs e)
        {
            SendKeys.Send("{DEL}");
        }

        private void OnClickContextMenuComputeArea(object sender, System.EventArgs e)
        {
            if (this._contextMenuShapes != null && this._contextMenuShapes.Count > 0)
            {
                foreach (IDrawable drawable in this._contextMenuShapes)
                {
                    if (drawable is ICircle)
                    {
                        double area = this._spatialDb.ComputeCircleArea(drawable);
                        // drawable must exist in the database
                        if (area == 0)
                        {
                            this._statusBar.Text = "The selected entity does not exist in the database.";
                        }
                        else
                        {
                            this._statusBar.Text = string.Format("Area of the circle is {0}", area.ToString());
                        }
                        break;
                    }
                }
            }
        }

#endregion Context Menu Handlers

#region Private Helper Methods

        /// <summary>
        /// This helper method displays a simple OK/Cancel dialog asking the user
        /// for some input. The caption of the dialog is set by the caller.
        /// </summary>
        /// <param name="caption">The dialog caption</param>
        /// <returns>A non-null value if the user hits OK; null on Cancel</returns>
        private string DisplayFindFormWithCaption(string caption)
        {
            string findText = string.Empty;
            using (FindForm findForm = new FindForm(caption))
            {
                DialogResult dialogResult = findForm.ShowDialog(this);
                if (dialogResult == DialogResult.OK)
                {
                    findText = findForm.FindText;
                }
                else if (dialogResult == DialogResult.Cancel)
                {
                    findText = null;
                }
            }

            return findText;
        }

        private void DisplayIndexStatistics()
        {
            if (this._index == null) 
                return;

            string indexNumbers = string.Format(
                                    @"Index Status: {0} levels, {1} nodes, {2} leaves",
                                    this._index.GetNumLevels(),
                                    this._index.GetNumNodes(),
                                    this._index.GetNumLeaves()
                                );
            this._statusBar.Text = indexNumbers;
        }

#endregion Private Helper Methods

#region IndexComparer

        /// <summary>
        /// A helper class that implements the IComparer interface used to sort
        /// the spatial index rectangles
        /// </summary>
        internal class IndexComparer : IComparer<IDrawable>
        {
            public int Compare(IDrawable shape1, IDrawable shape2)
            {
                Rectangle rectangle1 = shape1.GetBoundingBox();
                int area1 = rectangle1.Width * rectangle1.Height;

                Rectangle rectangle2 = shape2.GetBoundingBox();
                int area2 = rectangle2.Width * rectangle2.Height;

                if (area1 == area2) return 0;
                return area1 < area2 ? 1 : -1;  // deliberate: sort descending
            }
        }

#endregion IndexComparer

    }
}
