﻿using System;
using System.Drawing;
using System.Collections.Generic;

namespace SketchPad.Spatial
{
    /// <summary>
    /// A node in the spatial index
    /// </summary>
    public class IndexNode : IIndexNode
    {
        private IDrawable _shape;
        private Rectangle _boundingBox;
        private List<IndexNode> _children;

        public IndexNode()
        {
            this._shape = null;
            this._boundingBox = new Rectangle(Int32.MaxValue, Int32.MaxValue, Int32.MinValue, Int32.MinValue);
            this._children = null;
        }

        public IndexNode(IDrawable shape)
        {
            System.Diagnostics.Debug.Assert(shape != null);

            this._shape = shape;
            this._boundingBox = shape.GetBoundingBox();
            this._children = null;
        }

        public IDrawable Shape
        {
            get
            {
                return this._shape;
            }
        }

        public Rectangle BoundingBox
        {
            get
            {
                return this._boundingBox;
            }
        }

        public IList<IndexNode> Children
        {
            get
            {
                return this._children;
            }
        }

        public void AddChild(IndexNode node)
        {
            if (node == null)
            {
                return;
            }

            if (this._children == null)
            {
                this._children = new List<IndexNode>();
            }
            this._children.Add(node);
        }

        public void ExpandBoundingBox(Rectangle rectangle)
        {
            int topLeftX = Math.Min(this.BoundingBox.X, rectangle.X);
            int topLeftY = Math.Min(this.BoundingBox.Y, rectangle.Y);

            int width = Math.Max(this.BoundingBox.Right-topLeftX, rectangle.Right-topLeftX);
            int height = Math.Max(this.BoundingBox.Bottom-topLeftY, rectangle.Bottom-topLeftY);

            this._boundingBox = new Rectangle(topLeftX, topLeftY, width, height);
        }
    }

    /// <summary>
    /// The spatial index - esentially a pointer to the root node of the access
    /// structure.
    /// </summary>
    public class Index : ISpatialIndex
    {
        private IndexNode _root;

        public Index(IndexNode node)
        {
            System.Diagnostics.Debug.Assert(node != null);
            this._root = node;
        }

        public IndexNode Root
        {
            get
            {
                return this._root;
            }
        }

        public IDrawable GetShapeByLocation(Point pt, int ID)
        {
            return GetShapeByLocationHelper(pt, ID, this._root);
        }

        private IDrawable GetShapeByLocationHelper(Point pt, int ID, IndexNode node)
        {
            if (!this.PointInRectangle(pt, node.BoundingBox))
            {
                return null;
            }

            if (node.Children == null)
            {
                int nodeID = node.Shape is ILine ? ((ILine)node.Shape).ID : ((ICircle)node.Shape).ID;
                if (ID == nodeID) return node.Shape;
                return null;
            }

            System.Diagnostics.Debug.Assert(node.Children.Count >= 1);

            foreach (IndexNode child in node.Children)
            {
                IDrawable result = this.GetShapeByLocationHelper(pt, ID, child);
                if (result != null)
                    return result;  // no need to recurse further
            }
            return null;    // couldn't find anything
        }

        public IList<IDrawable> GetSelectedShapes(Rectangle rectangle)
        {
            return GetSelectedShapesHelper(rectangle, this._root);
        }

        private IList<IDrawable> GetSelectedShapesHelper(Rectangle rectangle, IndexNode node)
        {
            List<IDrawable> result = new List<IDrawable>();
            if (!rectangle.IntersectsWith(node.BoundingBox))
            {
                return result;
            }

            if (node.Children == null)
            {
                result.Add(node.Shape);
                return result;
            }

            System.Diagnostics.Debug.Assert(node.Children.Count >= 1);

            foreach (IndexNode child in node.Children)
            {
                result.AddRange(GetSelectedShapesHelper(rectangle, child));
            }
            return result;
        }

        public IList<IDrawable> GetShapesUnderPoint(Point pt)
        {
            return GetShapesUnderPointHelper(pt, this._root);
        }

        private IList<IDrawable> GetShapesUnderPointHelper(Point pt, IndexNode node)
        {
            List<IDrawable> result = new List<IDrawable>();

            if (!this.PointInRectangle(pt, node.BoundingBox))
            {
                return result;
            }

            if (node.Children == null)
            {
                result.Add(node.Shape);
                return result;
            }

            System.Diagnostics.Debug.Assert(node.Children.Count >= 1);

            foreach (IndexNode child in node.Children)
            {
                result.AddRange(GetShapesUnderPointHelper(pt, child));
            }
            return result;
        }

        /// <summary>
        /// If the point lies on the edge of the bounding box, Rectangle.Contains()
        /// returns false; so we need to compensate for that by testing against an
        /// appropriately inflated rectangle. This also takes care of zero-width and/or
        /// zero-height rectangles (i.e., ones that are in essence vertical or horzontal
        /// lines).
        /// </summary>
        /// <param name="pt">The point to test</param>
        /// <param name="rectangle">The rectangle against which to test the point</param>
        /// <returns>true if the point lies inside or on the boundary of the rectangle</returns>
        private bool PointInRectangle(Point pt, Rectangle rectangle)
        {
            bool checkInflatedRectangle = false;
            Rectangle inflated = rectangle;

            // check if the point lies on either horizontal edge of the rectangle
            if (pt.X == rectangle.Left || pt.X == rectangle.Right)
            {
                inflated.Inflate(1, 0);
                checkInflatedRectangle = true;
            }

            // check if the point lies on either vertical edge of the rectangle
            if (pt.Y == rectangle.Y || pt.Y == rectangle.Bottom)
            {
                inflated.Inflate(0, 1);
                checkInflatedRectangle = true;
            }

            if (checkInflatedRectangle)
            {
                if (inflated.Contains(pt))
                {
                    return true;
                }
            }
            else
            if (rectangle.Contains(pt))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Helper method to determine whether a point lies on a zero-width or zero-height
        /// rectangle (because in that situation, Rectangle.Contains() always returns false).
        /// </summary>
        /// <param name="pt">The Point to test</param>
        /// <param name="rectangle">The zero-width or zero-height rectangle to test against</param>
        /// <returns>true if the point lies on the linear rectangle, false otherwise</returns>
        private bool PointOnLine(Point pt, Rectangle rectangle)
        {
            System.Diagnostics.Debug.Assert(rectangle.Width == 0 || rectangle.Height == 0);

            if (rectangle.Width == 0 &&
                 pt.X == rectangle.Location.X &&
                 pt.Y >= rectangle.Location.Y &&
                 pt.Y <= rectangle.Location.Y + rectangle.Height)
            {
                return true;
            }
            if (rectangle.Height == 0 &&
                 pt.Y == rectangle.Location.Y &&
                 pt.X >= rectangle.Location.X &&
                 pt.X <= rectangle.Location.X + rectangle.Width)
            {
                return true;
            }

            return false;
        }

        public int GetNumLevels()
        {
            return GetNumLevelsHelper(this._root);
        }

        private int GetNumLevelsHelper(IndexNode node)
        {
            if (node.Children == null)
            {
                return 1;
            }

            int maxLevels = 0;
            foreach (IndexNode child in node.Children)
            {
                int numLevels = 1 + GetNumLevelsHelper(child);
                if (numLevels > maxLevels)
                    maxLevels = numLevels;
            }

            return maxLevels;
        }

        public int GetNumNodes()
        {
            return GetNumNodesHelper(this._root);
        }

        private int GetNumNodesHelper(IndexNode node)
        {
            if (node.Children == null)
            {
                return 1;
            }

            int numNodes = 1;   // for the current node
            foreach (IndexNode child in node.Children)
            {
                numNodes += GetNumNodesHelper(child);
            }

            return numNodes;
        }

        public int GetNumLeaves()
        {
            return GetNumLeavesHelper(this._root);
        }

        private int GetNumLeavesHelper(IndexNode node)
        {
            if (node.Children == null)
            {
                return 1;
            }

            int numLeaves = 0;
            foreach (IndexNode child in node.Children)
            {
                numLeaves += GetNumLeavesHelper(child);
            }

            return numLeaves;
        }
    }
}
