﻿
using System;
using System.Drawing;
using System.Collections.Generic;


namespace SketchPad.Spatial
{
    /// <summary>
    /// Interface for all drawable shapes
    /// </summary>
    public interface IDrawable
    {
        void Draw(Graphics g, Pen pen);
        void Select(Graphics g, Pen pen);
        Rectangle GetBoundingBox();
    }

    /// <summary>
    /// Interface for all identifiable shapes
    /// </summary>
    public interface IIdentity
    {
        int ID { get; }
        string Description { get; }
    }

    /// <summary>
    /// Line-specific interface
    /// </summary>
    public interface ILine : IDrawable, IIdentity
    {
    }

    /// <summary>
    /// Circle-specific interface
    /// </summary>
    public interface ICircle : IDrawable, IIdentity
    {
    }

    public interface IIndexNode
    {
        IDrawable Shape { get; }
        Rectangle BoundingBox { get; }
        IList<IndexNode> Children { get; }
        void AddChild(IndexNode node);
        void ExpandBoundingBox(Rectangle rectangle);
    }

    /// <summary>
    /// The spatial index over all shapes
    /// </summary>
    public interface ISpatialIndex
    {
        IndexNode Root { get; }
        IDrawable GetShapeByLocation(Point pt, int ID);
        IList<IDrawable> GetSelectedShapes(Rectangle rectangle);
        IList<IDrawable> GetShapesUnderPoint(Point pt);

        int GetNumLevels();
        int GetNumNodes();
        int GetNumLeaves();
    }

    /// <summary>
    /// A delegate that can be invoked during the construction of the
    /// spatial index
    /// </summary>
    /// <param name="rectangle">The rectangle that is being added to
    /// the index</param>
    public delegate void DisplayBoundingBox(Rectangle rectangle);

    /// <summary>
    /// Entry-point for spatial operations
    /// </summary>
    public abstract class SpatialDb : IDisposable
    {
        public abstract ILine CreateLine(Point A, Point B, string description);
        public abstract ICircle CreateCircle(Point center, Point circumference, string description);
        public abstract void DeleteShape(IDrawable shape);
        public abstract double ComputeCircleArea(IDrawable shape);
        public abstract IList<IDrawable> GetAllShapes();
        public abstract string GetAllDescriptions();
        public abstract IList<IDrawable> FindShapes(string description);
        public abstract void SaveChanges();

        public abstract DisplayBoundingBox DisplayBoundingBoxDelegate { get; set; }

        /// <summary>
        /// Axes used as the sorting dimension during index construction
        /// </summary>
        private enum SortAxis
        {
            X,
            Y
        }

        /// <summary>
        /// Build a BVH spatial index over the set of shapes specified in the argument
        /// </summary>
        /// <param name="shapes">The set of shapes to build an index over</param>
        /// <returns>The index object</returns>
        public ISpatialIndex BuildIndex(IList<IDrawable> shapes)
        {
            if (shapes == null || shapes.Count == 0)
            {
                return null;
            }

            IDrawable[] arrShapes = ((List<IDrawable>)shapes).ToArray();
            return new Index(BuildIndexHelper(arrShapes, 0, arrShapes.Length - 1, SortAxis.X));
        }

        private IndexNode BuildIndexHelper(IDrawable[] arrShapes, int first, int last, SortAxis sortAxis)
        {
            if (last < first)
            {
                System.Diagnostics.Debug.Assert(false);
                return null;
            }

            if (last == first)
            {
                IndexNode leaf = new IndexNode(arrShapes[first]);
                // ----- Display bounding box -----
                if (DisplayBoundingBoxDelegate != null)
                {
                    DisplayBoundingBoxDelegate(leaf.BoundingBox);
                }
                // --------------------------------
                return leaf;
            }

            IndexNode parent = new IndexNode();
            for (int i = first; i <= last; ++i)
            {
                parent.ExpandBoundingBox(arrShapes[i].GetBoundingBox());
            }

            // ----- Display bounding box -----
            if (DisplayBoundingBoxDelegate != null)
            {
                DisplayBoundingBoxDelegate(parent.BoundingBox);
            }
            // --------------------------------

            // sort (only) the slice of the array passed to us
            Array.Sort(arrShapes, first, last - first + 1, new Comparer(sortAxis));

            // flip the sort axis
            sortAxis = sortAxis == SortAxis.X ? SortAxis.Y : SortAxis.X;
            parent.AddChild(BuildIndexHelper(arrShapes, first, (first + last) / 2, sortAxis));
            parent.AddChild(BuildIndexHelper(arrShapes, (first + last) / 2 + 1, last, sortAxis));

            return parent;
        }

        private class Comparer : IComparer<IDrawable>
        {
            private SortAxis _sortAxis;
            internal Comparer(SortAxis sortAxis)
            {
                this._sortAxis = sortAxis;
            }
            public int Compare(IDrawable shape1, IDrawable shape2)
            {
                if (this._sortAxis == SortAxis.X) return Comparer.ComparisonX(shape1, shape2);
                return Comparer.ComparisonY(shape1, shape2);
            }
            private static int ComparisonX(IDrawable shape1, IDrawable shape2)
            {
                System.Diagnostics.Debug.Assert(shape1 != null);
                System.Diagnostics.Debug.Assert(shape2 != null);

                Rectangle rectangle1 = shape1.GetBoundingBox();
                Rectangle rectangle2 = shape2.GetBoundingBox();

                if (rectangle1.X == rectangle2.X) return 0;
                return rectangle1.X < rectangle2.X ? -1 : 1;
            }

            private static int ComparisonY(IDrawable shape1, IDrawable shape2)
            {
                System.Diagnostics.Debug.Assert(shape1 != null);
                System.Diagnostics.Debug.Assert(shape2 != null);

                Rectangle rectangle1 = shape1.GetBoundingBox();
                Rectangle rectangle2 = shape2.GetBoundingBox();

                if (rectangle1.Y == rectangle2.Y) return 0;
                return rectangle1.Y < rectangle2.Y ? -1 : 1;
            }
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected abstract void Dispose(bool disposing);
    }

    /// <summary>
    /// This is not a real entity (in the EDM/L2S sense); it has no database counterpart.
    /// It is used to display the bounding boxes created as part of the spatial index.
    /// </summary>
    public class IndexBox : IDrawable
    {
        private Rectangle _rectangle;
        public IndexBox(Rectangle rectangle)
        {
            this._rectangle = rectangle;
        }

        public void Draw(Graphics g, Pen pen)
        {
            g.DrawRectangle(pen, this._rectangle);
        }

        public void Select(Graphics g, Pen pen)
        {
            // no op
        }

        public Rectangle GetBoundingBox()
        {
            return _rectangle;
        }
    }
}